# Copyright © 2022 Ralph Seichter
#
# This file is part of LetsDNS.
#
# LetsDNS is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# LetsDNS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with LetsDNS.
# If not, see <https://www.gnu.org/licenses/>.
import socket
import struct
from unittest import skipUnless

from cryptography.x509 import BasicConstraints
from cryptography.x509 import Certificate
from dns.rdata import from_text
from dns.rdataclass import RdataClass
from dns.rdataset import Rdataset
from dns.rdatatype import RdataType

import tests
from letsdns.crypto import dane_tlsa_records
from letsdns.crypto import read_x509_cert
from letsdns.liveupdate import DnsLiveUpdate
from letsdns.tlsa import record_name
from tests import ENABLE_DEVELOPER_TESTS
from tests import ENABLE_LIVEUPDATE_TESTS


class TlsaTest(tests.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.update = DnsLiveUpdate()

    @classmethod
    @skipUnless(ENABLE_LIVEUPDATE_TESTS, 'online tests disabled')
    def tearDownClass(cls) -> None:
        super().tearDownClass()
        ds = Rdataset(RdataClass.IN, RdataType.TLSA, ttl=3)
        u = DnsLiveUpdate()
        u.execute(cls.c, name='_25._tcp', dataset=ds)
        u.execute(cls.c, name='test', dataset=ds)

    @skipUnless(ENABLE_LIVEUPDATE_TESTS, 'online tests disabled')
    def test_update_dns(self):
        self.c.active_section = 'tlsa'
        rd = from_text(RdataClass.IN, RdataType.TLSA, tok='3 1 1 1234')
        ds = Rdataset(RdataClass.IN, RdataType.TLSA, ttl=3)
        ds.add(rd)
        rc = self.update.execute(self.c, name='test', dataset=ds)
        self.assertEqual(0, rc)

    def test_bad_ttl(self):
        self.c.active_section = 'bad_ttl'
        rd = from_text(RdataClass.IN, RdataType.TLSA, tok='2 0 1 abcd')
        ds = Rdataset(RdataClass.IN, RdataType.TLSA, ttl=-3)
        ds.add(rd)
        with self.assertRaises(struct.error):
            self.update.execute(self.c, name='test', dataset=ds)

    def test_bad_nameserver(self):
        self.c.active_section = 'bad_ns'
        with self.assertRaises(socket.gaierror):
            self.update.execute(self.c, name='test', dataset=Rdataset(RdataClass.IN, RdataType.TLSA, ttl=3))

    def test_record_name(self):
        x = record_name(self.c, '98')
        h = self.c.get_mandatory('hostname')
        self.assertEqual(f'_98._tcp.{h}', x)

    def test_record_name_apex(self):
        self.c.active_section = 'apex'
        x = record_name(self.c, '76')
        self.assertEqual('_76._tcp', x)


@skipUnless(ENABLE_DEVELOPER_TESTS, 'developer tests disabled')
class CertTest(tests.TestCase):
    def _cert(self, name: str) -> Certificate:
        self.c.active_section = 'dane'
        return read_x509_cert(self.c.get_mandatory(name))

    def test_ca_cert(self):
        c = self._cert('cert_ca_path')
        b: BasicConstraints = c.extensions.get_extension_for_class(BasicConstraints).value
        self.assertTrue(b.ca)

    def test_ca_dane(self):
        c = self._cert('cert_ca_path')
        d = dane_tlsa_records(c)
        self.assertEqual('2 1 1 ', d[0][:6])

    def test_leaf_cert(self):
        c = self._cert('cert_leaf_path')
        b: BasicConstraints = c.extensions.get_extension_for_class(BasicConstraints).value
        self.assertFalse(b.ca)

    def test_leaf_dane(self):
        c = self._cert('cert_leaf_path')
        d = dane_tlsa_records(c)
        self.assertEqual('3 1 2 ', d[1][:6])


class ActionTest(tests.TestCase):
    @skipUnless(ENABLE_LIVEUPDATE_TESTS, 'online tests disabled')
    def test_tlsa(self):
        self.c.active_section = 'dane'
        DnsLiveUpdate.lifecycle(self.c, DnsLiveUpdate())



# Copyright © 2022 Ralph Seichter
#
# This file is part of LetsDNS.
#
# LetsDNS is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# LetsDNS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with LetsDNS.
# If not, see <https://www.gnu.org/licenses/>.
from unittest import skipUnless

from letsdns.hetznerapi import HetznerApiUpdate
from tests import ENABLE_HETZNER_API_TESTS
from tests import TestCase


class HetznerUpdateTest(TestCase):
    @skipUnless(ENABLE_HETZNER_API_TESTS, 'Hetzner API tests disabled')
    def test_lifecycle(self):
        self.c.active_section = 'hetzner'
        rc = HetznerApiUpdate.lifecycle(self.c, HetznerApiUpdate())
        self.assertEqual(0, rc)

    def test_empty_records(self):
        rc = HetznerApiUpdate().execute(self.c, records=[])
        self.assertEqual(0, rc)



# Copyright © 2022 Ralph Seichter
#
# This file is part of LetsDNS.
#
# LetsDNS is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# LetsDNS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with LetsDNS.
# If not, see <https://www.gnu.org/licenses/>.
from logging import debug

from letsdns.action import Action
from letsdns.configuration import Config


class AnAction(Action):
    def setup(self, conf: Config) -> int:
        r = super().setup(conf)
        debug('AnAction.setup')
        return r

    def execute(self, conf: Config, *args, **kwargs) -> int:
        return 0


class NotAnAction:
    def foo(self):
        pass



"""
Copyright © 2022 Ralph Seichter

This file is part of LetsDNS.

LetsDNS is free software: you can redistribute it and/or modify it under the terms of the GNU
General Public License as published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

LetsDNS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along with LetsDNS.
If not, see <https://www.gnu.org/licenses/>.
"""
import os
import unittest
from os import environ

from letsdns.configuration import Config
from letsdns.configuration import init_logger
from letsdns.util import is_truthy

ENABLE_DEVELOPER_TESTS = is_truthy(os.environ.get('ENABLE_DEVELOPER_TESTS'))
ENABLE_HETZNER_API_TESTS = is_truthy(os.environ.get('ENABLE_HETZNER_API_TESTS'))
ENABLE_LIVEUPDATE_TESTS = is_truthy(os.environ.get('ENABLE_LIVEUPDATE_TESTS'))
UNITTEST_CONF = 'UNITTEST_CONF'


def read_config(path: str = UNITTEST_CONF, active_section: str = 'DEFAULT') -> Config:
    if path == UNITTEST_CONF:
        path = environ.get(path)
        if path is None:  # pragma: no cover
            raise ValueError(f'Missing environment variable {UNITTEST_CONF}')
    conf = Config()
    conf.init(path)
    conf.active_section = active_section
    return conf


class TestCase(unittest.TestCase):
    c: Config

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        init_logger()
        TestCase.c = read_config()



# Copyright © 2022 Ralph Seichter
#
# This file is part of LetsDNS.
#
# LetsDNS is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# LetsDNS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with LetsDNS.
# If not, see <https://www.gnu.org/licenses/>.
from logging import debug
from logging import error
from typing import List
from typing import Optional

import requests
from requests import RequestException
from requests import Response

from letsdns.action import Action
from letsdns.configuration import Config
from letsdns.tlsa import record_name
from letsdns.tlsa import tlsa_records


def _content(response: Response) -> Optional[dict]:
    """Return API response content (i.e. JSON body data) in a dict. Debug the response data if necessary.
    May raise a RequestException in case of decoding errors.

    Args:
        response: Response object
    """
    debug(f'{response.status_code} {response.reason}')
    d: dict = response.json()
    if 'error' in d and isinstance(d['error'], list) and len(d['error']) > 0:
        # Non-empty error data found
        error(d['error'])
    return d


class HetznerApiUpdate(Action):
    @classmethod
    def lifecycle(cls, conf: Config, action: Action) -> int:
        """Process configuration data."""
        return action.execute(conf, records=tlsa_records(conf))

    def __init__(self) -> None:
        super().__init__()
        self.headers = {}
        self.api_token = ''
        self.api_url = 'https://dns.hetzner.com/api/v1'

    def zone_id(self, domain: str) -> Optional[str]:
        """Returns zone ID for the given domain name, or None in case of errors.

        Args:
            domain: Domain name, e.g. "example.com".
        """
        response = requests.get(f'{self.api_url}/zones', headers=self.headers, params={'name': domain})
        c = _content(response)
        if response.ok and 'zones' in c and len(c['zones']) > 0 and 'id' in c['zones'][0]:
            return c['zones'][0]['id']

    def find_records(self, zone_id: str, name: str) -> Optional[List[str]]:
        """Returns matching record ID list for the given zone ID and name, or None in case of errors.

        Args:
            zone_id: Parent zone ID.
            name: DNS record name.
        """
        response = requests.get(f'{self.api_url}/records', headers=self.headers, params={'zone_id': zone_id})
        c = _content(response)
        if response.ok and 'records' in c:
            id_list: List[str] = []
            name = name.lower()
            for record in c['records']:
                if 'name' in record and record['type'].lower() == 'tlsa' and record['name'].lower() == name:
                    id_list.append(record['id'])
            return id_list

    def cleanup_records(self, zone_id: str, name: str) -> bool:
        """Delete existing TLSA records. Returns True to indicate success.

        Args:
            zone_id: Parent zone ID.
            name: DNS record name.
        """
        id_list = self.find_records(zone_id, name)
        if id_list is None:  # pragma: no cover
            error('Error while attempting to find existing TLSA records')
            return False
        elif len(id_list) < 1:  # pragma: no cover
            debug('No existing TLSA records found')
            return True
        for record_id in id_list:
            response = requests.delete(f'{self.api_url}/records/{record_id}', headers=self.headers)
            # Debug response ignoring the actual content
            _content(response)
            if not response.ok:  # pragma: no cover
                return False
        return True

    def execute(self, conf: Config, *args, **kwargs) -> int:
        """Update DNS record using the Hetzner DNS API. Return 0 to indicate success."""
        records: List[str] = kwargs['records']
        if len(records) < 1:
            return 0
        self.api_token = conf.get_mandatory('api_token')
        self.headers = {'Auth-API-Token': self.api_token}
        self.api_url = conf.get('api_url', fallback=self.api_url)
        try:
            zone_id = self.zone_id(conf.get_mandatory('domain'))
            data = {
                'ttl': conf.get_ttl(),
                'type': 'TLSA',
                'zone_id': zone_id,
            }
            for port in conf.get_tcp_ports():
                name = record_name(conf, port)
                if zone_id is None or not self.cleanup_records(zone_id, name):  # pragma: no cover
                    error('Error while deleting existing TLSA records')
                    return 2
                for record in records:
                    data['name'] = name
                    data['value'] = record
                    response = requests.post(f'{self.api_url}/records', headers=self.headers, json=data)
                    # Debug response ignoring the actual content
                    _content(response)
                    if not response.ok:  # pragma: no cover
                        return 3
        except RequestException as e:  # pragma: no cover
            error(e)
            return 1
        return 0



# Copyright © 2022 Ralph Seichter
#
# This file is part of LetsDNS.
#
# LetsDNS is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# LetsDNS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with LetsDNS.
# If not, see <https://www.gnu.org/licenses/>.
import os.path
from unittest import skipUnless

from letsdns.action import Action
from letsdns.action import import_action
from letsdns.action import import_class
from letsdns.configuration import Config
from letsdns.core import action_class
from letsdns.core import traverse_config
from letsdns.liveupdate import DnsLiveUpdate
from letsdns.nsupdate import NsupdateStdout
from tests import ENABLE_LIVEUPDATE_TESTS
from tests import TestCase
from tests import read_config


class FailPre(Action):
    def setup(self, conf: Config) -> int:
        return -10

    def execute(self, conf: Config, *args, **kwargs) -> int:
        return 0

    def teardown(self, conf: Config) -> int:
        return 0


class FailExec(Action):
    def setup(self, conf: Config) -> int:
        return 0

    def execute(self, conf: Config, *args, **kwargs) -> int:
        return -20

    def teardown(self, conf: Config) -> int:
        return 0


class FailPost(Action):
    def setup(self, conf: Config) -> int:
        return 0

    def execute(self, conf: Config, *args, **kwargs) -> int:
        return 0

    def teardown(self, conf: Config) -> int:
        return -30


class ImportTest(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        ImportTest.c = read_config(path=os.path.join(os.path.dirname(__file__), 'actiontest.conf'))

    def test_import_missing_class(self):
        with self.assertRaises(ModuleNotFoundError):
            import_class('BAD.CLASS')

    def test_import_not_an_action(self):
        with self.assertRaises(TypeError):
            import_action('tests.actions.NotAnAction')

    def test_lookup_action(self):
        self.assertTrue(action_class('dane-tlsa') == DnsLiveUpdate)

    def test_unknown_action(self):
        self.assertIsNone(action_class('BAD'))

    def test_callable(self):
        self.assertIsNotNone(action_class('dynamic:tests.actions.AnAction'))

    def test_fail_pre(self):
        a = FailPre()
        self.assertEqual(-10, a.lifecycle(self.c, a))

    def test_fail_exec(self):
        a = FailExec()
        self.assertEqual(-20, a.lifecycle(self.c, a))

    def test_fail_post(self):
        a = FailPost()
        self.assertEqual(-30, a.lifecycle(self.c, a))

    def test_traverse(self):
        self.assertEqual(1, traverse_config(self.c))


class LiveUpdateTest(TestCase):
    @skipUnless(ENABLE_LIVEUPDATE_TESTS, 'online tests disabled')
    def test_lifecycle(self):
        self.c.active_section = 'dane'
        rc = DnsLiveUpdate.lifecycle(self.c, DnsLiveUpdate())
        self.assertEqual(0, rc)


class NsupdateTest(TestCase):
    def test_lifecycle(self):
        self.c.active_section = 'dane'
        rc = NsupdateStdout.lifecycle(self.c, NsupdateStdout())
        self.assertEqual(0, rc)



# Copyright © 2022 Ralph Seichter
#
# This file is part of LetsDNS.
#
# LetsDNS is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# LetsDNS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with LetsDNS.
# If not, see <https://www.gnu.org/licenses/>.
import logging
import os
import subprocess
from configparser import NoOptionError
from tempfile import NamedTemporaryFile

import tests
from letsdns.configuration import Config
from letsdns.configuration import log_level


class ConfigurationTest(tests.TestCase):
    def test_loglevel_bad(self):
        os.environ['LOG_LEVEL'] = 'BAD'
        with self.assertRaises(AttributeError):
            log_level()

    def test_loglevel_good(self):
        os.environ['LOG_LEVEL'] = 'FATAL'
        self.assertEqual(logging.FATAL, log_level())

    def test_dump(self):
        conf = Config()
        conf.init()
        f = NamedTemporaryFile(mode='wt', delete=False)
        conf.dump(f)
        f.close()
        diff = subprocess.run(['diff', 'dump-expected', f.name])
        os.unlink(f.name)
        self.assertEqual(0, diff.returncode)

    def test_get_domain(self):
        self.c.active_section = 'DEFAULT'
        with self.assertRaises(NoOptionError):
            self.c.get_domain()

    def test_options(self):
        self.c.active_section = 'tlsa'
        self.assertGreater(len(self.c.options()), 5)



# Copyright © 2022 Ralph Seichter
#
# This file is part of LetsDNS.
#
# LetsDNS is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# LetsDNS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with LetsDNS.
# If not, see <https://www.gnu.org/licenses/>.
import logging
import sys
from configparser import ConfigParser
from configparser import ExtendedInterpolation
from logging import basicConfig
from logging import debug
from typing import List

from letsdns.util import getenv
from letsdns.util import is_sensitive
from letsdns.util import split

LOG_LEVEL = 'LOG_LEVEL'


def log_level(default: str = 'ERROR') -> int:
    """Return logging level, using the LOG_LEVEL environment variable if available.

    Args:
        default: The default log level.
    """
    return getattr(logging, getenv(LOG_LEVEL, default, debug_env=False).upper())


def init_logger() -> None:
    """Initialise logger."""
    try:
        level = log_level()
        basicConfig(datefmt='%Y-%m-%d %H:%M:%S', format='%(asctime)s %(levelname)s %(message)s', level=level)
    except AttributeError as e:  # pragma: no cover
        print(e, file=sys.stderr)
        sys.exit(1)


class Config:
    """Provide access to configuration data."""
    parser: ConfigParser
    active_section: str  # The currently active configuration section

    def dump(self, destination=sys.stdout) -> None:
        """Dump configuration state into a file.

        Args:
            destination: File pointer.
        """
        self.parser.write(destination)

    def init(self, filenames=None) -> None:
        """Initialise object by loading configuration files from disk.
        Nonexisting or unreadable files are silently ignored.

        Args:
            filenames: Either a single string or a list of strings.
        """
        self.parser = ConfigParser(interpolation=ExtendedInterpolation())
        self.parser.read_dict({'DEFAULT': {'nameserver': '127.0.0.1'}})
        if filenames:
            self.parser.read(filenames, encoding='utf-8')

    def get(self, name: str, fallback=None) -> str:
        """Return an optional configuration value or the specified fallback value.

        Args:
            name: Option name.
            fallback: Returned if option is undefined.
        """
        return self.parser.get(self.active_section, name, fallback=fallback)

    def get_mandatory(self, name: str) -> str:
        """Return a mandatory configuration value.
        Raise an exception if option value is undefined.

        Args:
            name: Option name.
        """
        v = self.parser.get(self.active_section, name)
        if is_sensitive(name):
            debug(f'config: {name} = *****')
        else:
            debug(f'config: {name} = {v}')
        return v

    def get_domain(self) -> str:
        """Return the mandatory 'domain' configuration value.

        Raise an exception if 'domain' is undefined.
        """
        return self.get_mandatory('domain')

    def get_ttl(self) -> int:
        """Return the 'ttl' configuration value (seconds, integer)."""
        return int(self.get('ttl', fallback='1800'))

    def get_tcp_ports(self) -> List[str]:
        """Return the optional 'tcp_ports' configuration value."""
        return split(self.get('tcp_ports', fallback='25'))

    def options(self) -> List[str]:
        """Return all options in the active section."""
        return self.parser.options(self.active_section)



# Copyright © 2022 Ralph Seichter
#
# This file is part of LetsDNS.
#
# LetsDNS is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# LetsDNS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with LetsDNS.
# If not, see <https://www.gnu.org/licenses/>.
from argparse import ArgumentParser
from logging import debug

from letsdns import HOMEPAGE
from letsdns import IDENTIFIER
from letsdns import VERSION
from letsdns.configuration import Config
from letsdns.configuration import init_logger
from letsdns.core import traverse_config


def main():
    init_logger()
    parser = ArgumentParser(
        description=f'LetsDNS {VERSION} - Manage DANE TLSA records in DNS servers.',
        epilog=f'See {HOMEPAGE} for more information.',
        fromfile_prefix_chars='@',
        prog=IDENTIFIER,
    )
    parser.add_argument('--showconfig', action='store_true', help='print effective configuration to stdout')
    parser.add_argument('configfile', type=str, nargs='+')
    args = parser.parse_args()
    config = Config()
    config.init(args.configfile)
    if args.showconfig:
        config.dump()
    else:
        n = traverse_config(config)
        debug(f'Processed {n} action(s)')


if __name__ == '__main__':
    main()



# Copyright © 2022 Ralph Seichter
#
# This file is part of LetsDNS.
#
# LetsDNS is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# LetsDNS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with LetsDNS.
# If not, see <https://www.gnu.org/licenses/>.
import json
from _socket import gethostbyname
from logging import debug

from dns import query
from dns import rcode
from dns import tsigkeyring
from dns.message import Message
from dns.update import Update

from letsdns.action import Action
from letsdns.configuration import Config
from letsdns.tlsa import rdata_action_lifecycle
from letsdns.tlsa import record_name
from letsdns.util import getenv


class DnsLiveUpdate(Action):
    @classmethod
    def lifecycle(cls, conf: Config, action: Action) -> int:
        return rdata_action_lifecycle(conf, action)

    def execute(self, conf: Config, *args, **kwargs) -> int:
        """Update DNS record using the dnspython library. Return 0 to indicate success."""
        zone = conf.get_mandatory('domain')
        path = conf.get('keyfile')
        if path:
            with open(path, 'r') as f:
                keyring = tsigkeyring.from_text(json.load(f))
        else:  # pragma: no cover
            keyring = None
        update = Update(zone=zone, keyring=keyring)
        for port in conf.get_tcp_ports():
            name = record_name(conf, port)
            update.delete(name)
            dataset = kwargs['dataset']
            if len(dataset) > 0:
                update.replace(name, dataset)
            nameserver = gethostbyname(conf.get_mandatory('nameserver'))
            debug(f'DNS update: {update}')
            t = int(getenv('DNS_TIMEOUT_SECONDS', '30'))
            response: Message = query.tcp(update, nameserver, timeout=t)
            debug(f'DNS Response: {response}')
            if rcode.from_flags(response.flags, 0) != rcode.NOERROR:  # pragma: no cover
                return 1
        return 0



# Copyright © 2022 Ralph Seichter
#
# This file is part of LetsDNS.
#
# LetsDNS is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# LetsDNS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with LetsDNS.
# If not, see <https://www.gnu.org/licenses/>.
import os
import re
from logging import debug
from typing import List

_SENSITIVE_RE = re.compile(r'key|pass|token', re.IGNORECASE)


def is_sensitive(name: str) -> bool:
    """Return True if the specified name is considered sensitive."""
    return _SENSITIVE_RE.search(name) is not None


def is_truthy(something) -> bool:
    """Test if 'something' represents boolean True.

    An object represents True if it is (or can be converted to) a string and said
    string's first character, converted to lowercase, is one of: 1, t, y.

    Args:
        something: The object to examine.
    """
    if not something:
        return False
    if isinstance(something, str):
        s = something
    else:
        s = str(something)
    f = s[:1].lower()
    t = '1ty'.find(f)
    return t >= 0


def getenv(name: str, default=None, debug_env=True):
    """Return environment variable value if available, otherwise return the default value.
    Args:
        name: Environment variable name.
        default: Default value.
        debug_env: Debug variable name/value.
    """
    if name in os.environ:
        value = os.environ[name]
    else:
        value = default
    if debug_env:
        if is_sensitive(name):
            debug(f'ENV {name}=*****')
        else:
            debug(f'ENV {name}={value}')
    return value


def split(source: str) -> List[str]:
    """Split a string where whitespace or comma occurs. Return empty list for invalid input.

    Args:
        source: The string to be split.
    """
    strings = []
    if source:
        for s in source.replace(',', ' ').split(' '):
            strings.append(s.strip())
    return list(filter(lambda x: x, strings))



"""
Copyright © 2022 Ralph Seichter

This file is part of LetsDNS.

LetsDNS is free software: you can redistribute it and/or modify it under the terms of the GNU
General Public License as published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

LetsDNS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along with LetsDNS.
If not, see <https://www.gnu.org/licenses/>.
"""
HOMEPAGE = 'https://letsdns.org/'
IDENTIFIER = 'letsdns'
VERSION = '1.0'



# Copyright © 2022 Ralph Seichter
#
# This file is part of LetsDNS.
#
# LetsDNS is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# LetsDNS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with LetsDNS.
# If not, see <https://www.gnu.org/licenses/>.
import uuid
from os import getenv

import tests
from letsdns.util import is_truthy


class UtilTest(tests.TestCase):
    def test_getenv(self):
        u = uuid.uuid4().hex
        self.assertIsNone(getenv(u))
        self.assertEqual(-12, getenv(u, default=-12))

    def test_sensitive(self):
        u = 'API_TOKEN_' + uuid.uuid4().hex
        self.assertIsNone(getenv(u))

    def test_is_truthy1(self):
        self.assertFalse(is_truthy('false'))

    def test_is_truthy2(self):
        self.assertFalse(is_truthy(None))

    def test_is_truthy3(self):
        self.assertTrue(is_truthy('YES'))

    def test_is_truthy4(self):
        self.assertTrue(is_truthy(1))



# Copyright © 2022 Ralph Seichter
#
# This file is part of LetsDNS.
#
# LetsDNS is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# LetsDNS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with LetsDNS.
# If not, see <https://www.gnu.org/licenses/>.
from logging import info
from typing import Optional

from letsdns.action import Action
from letsdns.action import import_action
from letsdns.configuration import Config
from letsdns.hetznerapi import HetznerApiUpdate
from letsdns.liveupdate import DnsLiveUpdate
from letsdns.nsupdate import NsupdateStdout

_ACTION_CLASS_MAP = {
    # Read-only mapping from action name to action class
    'dane-tlsa': DnsLiveUpdate,
    'hetzner-tlsa': HetznerApiUpdate,
    'nsupdate-stdout': NsupdateStdout,
}


def action_class(name: str) -> Optional[Action]:
    """Return action class for the given name.

    Args:
        name: Action name.
    """
    if name in _ACTION_CLASS_MAP:
        _class = _ACTION_CLASS_MAP[name]
    elif name.startswith('dynamic:'):
        _class = import_action(name.split(':')[1])
    else:
        _class = None
    return _class


def traverse_config(conf: Config) -> int:
    """Traverse and process the configuration sections. Return the number of processed actions.

    Args:
        conf: Configuration object.
    """
    counter = 0
    for section in conf.parser.sections():
        conf.active_section = section
        info(f'Config section: {section}')
        name = conf.get('action')
        if name:
            _class = action_class(name)
            if _class:
                description = f' {_class}.lifecycle'
                info(f'Start{description}')
                # noinspection PyCallingNonCallable
                _class.lifecycle(conf, _class())
                info(f'End{description}')
                counter += 1
    return counter



# Copyright © 2022 Ralph Seichter
#
# This file is part of LetsDNS.
#
# LetsDNS is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# LetsDNS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with LetsDNS.
# If not, see <https://www.gnu.org/licenses/>.
from abc import ABC
from abc import abstractmethod
from logging import debug
from logging import warning
from typing import Type

from letsdns.configuration import Config


class Action(ABC):
    """Abstract base class for LetsDNS actions."""

    @classmethod
    def lifecycle(cls, conf: Config, action) -> int:
        """Invoke the lifecycle methods of a dynamically imported action. The first non-zero
        method return code will abort execution and will be returned to the caller.

        Args:
            conf: Configuration data.
            action: Action object. Ignored in the default implementation.
        """

        def _log_rc(method):
            warning(f'{cls.__name__}.{method} returned code {rc}')

        rc = action.setup(conf)
        if rc == 0:
            rc = action.execute(conf)
            if rc == 0:
                rc = action.teardown(conf)
                if rc != 0:
                    _log_rc('teardown')
            else:
                _log_rc('execute')
        else:
            _log_rc('setup')
        return rc

    def setup(self, conf: Config) -> int:
        """Pre-execution phase, for initialisation."""
        debug(f'{self}.setup({conf})')
        return 0

    @abstractmethod
    def execute(self, conf: Config, *args, **kwargs) -> int:
        """Action classes MUST implement this method. Unknown positional and keyword arguments MUST be ignored.

        Args:
            conf: Configuration data.
            args: Positional arguments.
            kwargs: Keyword arguments.
        """
        raise NotImplementedError  # pragma: no cover

    def teardown(self, conf: Config) -> int:
        """Post-execution phase, for cleanup."""
        debug(f'{self}.teardown({conf})')
        return 0


def import_class(class_name: str):
    """Dynamically import a Python class. The containing module must be available in PYTHONPATH.

    Args:
        class_name: Fully qualified class name, e.g. "mymodule.submodule.MyClass".
    """
    debug(f'Import {class_name}')
    # Split fully qualified name into components, the last being the class name.
    components = class_name.split('.')
    # Import the module containing the class.
    attribute = __import__('.'.join(components[:-1]))
    # Traverse the attribute hierarchy.
    for component in components[1:]:
        attribute = getattr(attribute, component)
    return attribute


def import_action(class_name: str) -> Type[Action]:
    """Dynamically import an action class. Raises a TypeError exception if the imported
    class is not a subclass of 'Action'.

    Args:
        class_name: Fully qualified class name, e.g. "mymodule.actions.SomeAction".
    """
    class_ = import_class(class_name)
    if issubclass(class_, Action):
        return class_
    raise TypeError(f'{class_} is no subclass of {Action}')



# Copyright © 2022 Ralph Seichter
#
# This file is part of LetsDNS.
#
# LetsDNS is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# LetsDNS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with LetsDNS.
# If not, see <https://www.gnu.org/licenses/>.
import hashlib
from typing import List

# noinspection PyProtectedMember
from cryptography.hazmat.primitives._serialization import Encoding
# noinspection PyProtectedMember
from cryptography.hazmat.primitives._serialization import PublicFormat
from cryptography.x509 import BasicConstraints
from cryptography.x509 import Certificate
from cryptography.x509 import load_pem_x509_certificate


def read_x509_cert(filename: str) -> Certificate:
    """Read x509 certificate from file."""
    with open(filename, 'rb') as f:
        return load_pem_x509_certificate(f.read())


def sha_digests(something):
    """Generate hexadecimal SHA256 and SHA512 hashes for some data."""
    sha256 = hashlib.sha256()
    sha256.update(something)
    sha512 = hashlib.sha512()
    sha512.update(something)
    return sha256.hexdigest(), sha512.hexdigest()


def dane_tlsa_records(cert: Certificate) -> List[str]:
    """Return list of TLSA record data for the certificate.

    Args:
        cert: x509 certificate.
    """
    bc: BasicConstraints = cert.extensions.get_extension_for_class(BasicConstraints).value
    if bc.ca:
        usage = 2  # DANE-TA
    else:
        usage = 3  # DANE-EE
    pk = cert.public_key().public_bytes(format=PublicFormat.SubjectPublicKeyInfo, encoding=Encoding.DER)
    sha256, sha512 = sha_digests(pk)
    return [
        f'{usage} 1 1 {sha256}',
        f'{usage} 1 2 {sha512}'
    ]



# Copyright © 2022 Ralph Seichter
#
# This file is part of LetsDNS.
#
# LetsDNS is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# LetsDNS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with LetsDNS.
# If not, see <https://www.gnu.org/licenses/>.
from letsdns.action import Action
from letsdns.configuration import Config
from letsdns.tlsa import rdata_action_lifecycle
from letsdns.tlsa import record_name


class NsupdateStdout(Action):
    @classmethod
    def lifecycle(cls, conf: Config, action) -> int:
        return rdata_action_lifecycle(conf, action)

    def execute(self, conf: Config, *args, **kwargs) -> int:
        """Generate 'nsupdate' command list and print to stdout."""
        dataset = kwargs['dataset']
        nameserver = conf.get_mandatory('nameserver')
        ttl = int(conf.get_mandatory('ttl'))
        zone = conf.get_mandatory('domain')
        print(f'server {nameserver}')
        for port in conf.get_tcp_ports():
            name = record_name(conf, port)
            print(f'update del {name}.{zone}.', 'IN TLSA')
            for data in dataset:
                print(f'update add {name}.{zone}.', ttl, 'IN TLSA', data)
        print('send')
        return 0



# Copyright © 2022 Ralph Seichter
#
# This file is part of LetsDNS.
#
# LetsDNS is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# LetsDNS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with LetsDNS.
# If not, see <https://www.gnu.org/licenses/>.
import re
from typing import List

from dns.rdata import from_text
from dns.rdataclass import RdataClass
from dns.rdataset import Rdataset
from dns.rdatatype import RdataType

from letsdns.action import Action
from letsdns.configuration import Config
from letsdns.crypto import dane_tlsa_records
from letsdns.crypto import read_x509_cert


def record_name(conf: Config, tcp_port: str) -> str:
    """Return TLSA record name for the configured host name.
    A single dot '.' host name denotes the domain apex.

    Args:
        conf: Configuration object.
        tcp_port: Desired TCP service port, default 25.
    """
    h = conf.get_mandatory('hostname')
    if h == '.':
        suffix = ''
    else:
        suffix = f'.{h}'
    return f'_{tcp_port}._tcp{suffix}'


def tlsa_records(conf: Config) -> List[str]:
    """Generate list of TLSA record strings based on the given configuration options.

    Args:
        conf: Configuration object.
    """
    path_re = re.compile(r'^cert_\S+_path$')
    records: List[str] = []
    for option in conf.options():
        if path_re.match(option):
            path = conf.get_mandatory(option)
            cert = read_x509_cert(path)
            for record in dane_tlsa_records(cert):
                if record not in records:
                    records.append(record)
    return records


def rdata_action_lifecycle(conf: Config, action: Action) -> int:
    """Lifecycle method for Rdata-based actions."""
    records = tlsa_records(conf)
    if len(records) < 1:  # pragma: no cover
        return 0
    dataset = Rdataset(RdataClass.IN, RdataType.TLSA, ttl=conf.get_ttl())
    for record in records:
        dataset.add(from_text(RdataClass.IN, RdataType.TLSA, record))
    return action.execute(conf, dataset=dataset)
